/*****************************************************************************
* 
* check_perfmon.c - A Nagios plugin for fetching windows performance log data
*
* License: GPL
* Copyright (c) 2009 Thomas Guyot-Sionnest <tguyot@gmail.com>
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*
*
*****************************************************************************/

const char *progname = "check_perfmon";
const char *copyright = "2009";
const char *email = "tguyot@gmail.com";
/* This is meant to be concatenated with NP_VERSION */
const char *version = "0.9/nagios-plugins-v";

#include "common.h"
#include "utils.h"
#include "utils_base.h"
#include "getlog_base.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifndef NP_GETLOG_READ_CHNK
#define NP_GETLOG_READ_CHNK 512 * 2 /* 1024 bytes */
#endif
#ifndef NP_GETLOG_MAX_READ
#define NP_GETLOG_MAX_READ 1024 * 512 /* 512KiB */
#endif

/* Own prototypes */
void getlog_exit(char *);
void perfmon_exit(char *message, char *errstr);
int process_arguments(int, char **);
void print_help (void);
void print_usage (void);

/* Global variables */
char *np_version = NULL;
char *hostname = NULL;
char *logdir = NULL;
char *counter = NULL;
char *perfdata_str = NULL;
int perminute = 0;
unsigned int max_age = 0;
char *warn_threshold = NULL;
char *crit_threshold = NULL;

int verbose=0;



int
main(int argc, char **argv)
{
	char *log = NULL;
	char *header, *last, *datestr, *value;
	char *perfdata_line = NULL;
	char *tmp = NULL;
	int result = STATE_UNKNOWN;
	int fd, index, diff, last_ts;
	double dvalue, last_value;
	char sep = '\\';
	int return_code=STATE_OK;
	thresholds *thresholds = NULL;
	struct timeval tv;

	setlocale (LC_ALL, "");
	bindtextdomain (PACKAGE, LOCALEDIR);
	textdomain (PACKAGE);

	asprintf(&np_version, "%s%s", version, NP_VERSION);

	/* Parse extra opts if any */
	argv=np_extra_opts(&argc, argv, progname);

	if(process_arguments(argc,argv)==ERROR)
		usage(_("Could not parse arguments"));

	/* Initialize the thresholds */
	set_thresholds(&thresholds, warn_threshold, crit_threshold);
	if(verbose)
		print_thresholds(progname, thresholds);

	/* Set externs to be used by getlog_base */
	read_chnk = NP_GETLOG_READ_CHNK;
	max_read = NP_GETLOG_MAX_READ;
	getlog_err = &getlog_exit;

	/* Open the log and get the first and last lines */
	log = malloc(strlen(logdir)+strlen(hostname)+6);
	sprintf(log, "%s/%s.csv", logdir, hostname);

	if ((fd = open(log, O_RDONLY)) == -1)
		perfmon_exit(_("Error opening log file"), strerror(errno));

	if ((header = get_head(fd)) == NULL)
		perfmon_exit(_("Failed to read first log line"), NULL);

	if ((last = get_tail(fd)) == NULL)
		perfmon_exit(_("Failed to read last log line"), NULL);

	close(fd);

	/* Strip the first column (we'll do the same to get the date field) */
	subst_col(0, &header);

	if (strcmp(counter, "list") == 0) {
		printf("Available counters:\n");
		while ((tmp = subst_col(0, &header)) != NULL) {
			tmp+= 2;
			if ((tmp = strchr(tmp, '\\')) == NULL)
				perfmon_exit(_("Invalid column header"), NULL);

			printf("%s\n", tmp);
		}
		exit(STATE_OK);
	}

	/* Check the date diff */
	datestr = subst_col(0, &last);
	if (verbose) printf("last timestamp: %s\n", datestr);
	if (max_age > 0) {
		if ((diff = datediff(datestr)) == -1)
			perfmon_exit(_("Couldn't parse date string"), datestr);

		if (verbose) printf("checking if %i > %i\n", diff, max_age);
		if (diff > max_age) {
			asprintf(&tmp, _("Data too old (%i seconds)"), diff);
			perfmon_exit(tmp, NULL);
		}
	}

	/*If the first character is not backslash, assume it's the separator and
	 * substitute the counter accordingly */
	if (strlen(counter) && counter[0] != sep) {
		sep = counter[0];
		tmp = counter;
		while (*tmp != '\0') {
			if (*tmp == sep) *tmp = '\\';
			tmp++;
		}
	}

	/* Get the index... */
	index = find_index(counter, header);
	if (verbose) printf("index: %i\n", index);

	/* Substitute backslashes (\) with carets (^) or the separator used, to print
	 * properly in Nagios. */
	if (sep == '\\')
		sep = '^';

	tmp = counter;
	while (*tmp != '\0') {
		if (*tmp == '\\') *tmp = sep;
		tmp++;
	}

	/* Now check the results of the find_index call above */
	if (index == -1) {
		asprintf(&tmp, _("No matching index '%s'"), counter);
		perfmon_exit(tmp, NULL);
	}

	/* then get the column */
	if ((value = subst_col(index, &last)) == NULL) {
		asprintf(&tmp, _("Column %i not found"), index);
		perfmon_exit(tmp, NULL);
	}
	if (verbose) printf("value: %s\n", value);

	if (!is_numeric(value))
		perfmon_exit(_("Value is not numeric"), value);
	dvalue = strtod(value, NULL);

	/* If we use perfdata, treat the value as a counter and return the average */
	perfdata_line = strdup("|");
	if (perfdata_str) {
		if (gettimeofday(&tv, NULL) == -1)
			perfmon_exit("gettimeofday", strerror(errno));

		asprintf(&perfdata_line, "%s %s", perfdata_line,
		         perfdata("time", tv.tv_sec, "", FALSE, 0, FALSE, 0, FALSE, 0, FALSE, 0));
		asprintf(&perfdata_line, "%s %s", perfdata_line,
		         fperfdata("last", dvalue, "", FALSE, 0, FALSE, 0, FALSE, 0, FALSE, 0));

		if (verbose)
			printf("Parsing perfdata string .%s.\n", perfdata_str);

		if (np_iperfdata(perfdata_str, "time", &last_ts) != NP_PERFDATA_OK ||
		    np_dperfdata(perfdata_str, "last", &last_value) != NP_PERFDATA_OK) {
			printf("PERFMON UNKNOWN: %s%s\n", _("Couldn't parse last counter (normal on first run)"), perfdata_line);
			exit(STATE_UNKNOWN);
		}

		/* Can't divide by 0 */
		if (tv.tv_sec == last_ts) {
			printf("PERFMON UNKNOWN: %s%s\n", _("Last and current times are the same"), perfdata_line);
			exit(STATE_UNKNOWN);
		}

		dvalue = (dvalue - last_value) / (tv.tv_sec - last_ts);
		if (perminute) dvalue *= 60;
	}

	if (verbose)
		printf("Retrieved value: %.10e\n", dvalue);

	/* Add perfdata */
	asprintf(&perfdata_line, "%s %s", perfdata_line,
	         fperfdata("counter", dvalue, "c",
	                   warn_threshold ? TRUE : FALSE, warn_threshold ? thresholds->warning->end : 0,
	                   crit_threshold ? TRUE : FALSE, crit_threshold ? thresholds->critical->end : 0,
	                   FALSE, 0, FALSE, 0));

	result = get_status(dvalue, thresholds);
	asprintf(&value, "%.2f", dvalue);
	/* Truncate the string of there' no decimal */
	if (strlen(value) >= 4 && strcmp(&value[strlen(value)-3], ".00") == 0)
		value[strlen(value)-3] = '\0';
	printf("PERFMON %s: %s: %s%s\n", state_text(result), counter, value, perfdata_line);
	exit(result);

}



/* Error handler for getlog_base */
void
getlog_exit(char *errstr)
{
	perfmon_exit(_("Perfmon log file error"), errstr);
}

/* General error handler */
void
perfmon_exit(char *message, char *errstr)
{

	if (errstr)
		printf("PERFMON UNKNOWN: %s: %s\n", message, errstr);
	else
		printf("PERFMON UNKNOWN: %s\n", message, errstr);

  exit(STATE_UNKNOWN);
}


int
process_arguments(int argc, char **argv)
{
	int c;
	int option=0;
	static struct option longopts[]= {
		{"hostname",  required_argument, 0,'H'},
		{"logdir",    required_argument, 0,'L'},
		{"counter",   required_argument, 0,'C'},
		{"warning",   required_argument, 0,'w'},
		{"critical",  required_argument, 0,'c'},
		{"perfdata",  required_argument, 0,'P'},
		{"maxage",    required_argument, 0,'M'},
		{"perminute", no_argument,       0,'m'},
		{"verbose",   no_argument,       0,'v'},
		{"version",   no_argument,       0,'V'},
		{"help",      no_argument,       0,'h'},
		{0,0,0,0}
	};

	/* no options were supplied */
	if(argc<2)
		return ERROR;

	while(1) {

		c = getopt_long(argc, argv, "hvVH:L:C:w:c:P:M:m", longopts, &option);

		if (c==-1 || c==EOF || c==1)
			break;

		switch(c) {

		case 'H': /* Hostname (log file basename) */
			hostname = optarg;
			break;

		case 'L': /* Log directory (<logdir>/<hostname>.csv) */
			logdir = optarg;
			break;

		case 'C': /* Counter */
			counter = optarg;
			break;

		case 'w': /* Warning Thresholds */
			warn_threshold = optarg;
			break;

		case 'c': /* Critical Thresholds */
			crit_threshold = optarg;
			break;

		case 'P': /* Performance data (state retention) */
			perfdata_str = optarg;
			break;

		case 'M': /* Max age */
			if (!is_intnonneg(optarg))
				usage2(_("t be a positive integer"), optarg);
			else
				max_age = atoi(optarg);
			break;

		case 'm': /* Make average data per-minute */
			perminute++;
			break;

		case 'V': /* version */
			print_revision(progname, np_version);
			exit(STATE_OK);
			break;

		case 'h': /* help */
			print_help();
			exit(STATE_UNKNOWN);
			break;

		case 'v': /* verbose */
			verbose++;
			break;

		default:
			return STATE_UNKNOWN;
			break;
		}
	}

	if (hostname == NULL)
		usage("hostname Required");
	if (logdir == NULL)
		usage("logdir required");
	if (counter == NULL)
		usage("counter required");

	return OK;
}

void
print_help(void)
{
	print_revision(progname, np_version);
	printf ("Copyright (c) 2009 Thomas Guyot-Sionnest (tguyot@gmail.com)\n\n");

	printf("   %s\n", _("Checks a Windows Performance log in CSV format. The file path is constructed as"));
	printf("   %s\n",_("<log_dir>/<hostname>.csv, case-sensitive."));
	printf("\n\n");

	print_usage();

	printf("\n");
	printf("%s\n", _("Options:"));
	printf(UT_EXTRA_OPTS);
	printf(" %s\n", "-H, --hostname=<hostname>");
	printf("    %s\n", _("Hostname to check (Log file base name)"));
	printf(" %s\n", "-L, --logdir=<log_directory>");
	printf("    %s\n", _("Directory location for log file names"));
	printf(" %s\n", "-C, --counter=<counter>|list");
	printf("    %s\n", _("Performance counter name. Use 'list' to get a list of available counters."));
	printf("    %s\n", _("If the counter do not begin by a backslash (\\) the first character will be"));
	printf("    %s\n", _("used as the separator (every occurrence of it will be replaced by a"));
	printf("    %s\n", _("backslash). The separator used will be printed in output. If the separator"));
	printf("    %s\n", _("is a backslash, it will always be replaced with a caret (^)."));
	printf(" %s\n", "-w, --warning=THRESHOLD");
	printf("    %s\n", _("Warning threshold for the counter"));
	printf(" %s\n", "-c, --critical=THRESHOLD");
	printf("    %s\n", _("Critical threshold for the counter"));
	printf(" %s\n", "-P, --perfdata=<string>");
	printf("    %s\n", _("Pass the current counter value and timestamp as additional performance data"));
	printf("    %s\n", _("and use it to get the 1-second average during the interval. You should pass"));
	printf("    %s\n", _("'$SERVICEPERFDATA$' as argument of this option (Don't forget to quote it!)."));
	printf(" %s\n", "-m, --perminute");
	printf("    %s\n", _("Only effective with -P, --perfdata; make the average per-minute. Also"));
	printf("    %s\n", _("affects thresholds (They will be against the per-minute average)."));
	printf(" %s\n", "-M, --maxage=<max_age>");
	printf("    %s\n", _("Maximum age in seconds of the last CSV entry for the data to be considered"));
	printf("    %s\n", _("valid, after which the plugin will return UNKNOWN."));
	printf("    %s\n", _("Default is no checking (0)"));

	printf(UT_VERBOSE);

	printf("\n");
	printf("%s\n", _("Notes:"));
	printf(UT_THRESHOLDS_NOTES);
#ifdef NP_EXTRA_OPTS
	printf("\n");
	printf(UT_EXTRA_OPTS_NOTES);
#endif

	printf("\n");
	printf(UT_SUPPORT);
}


void
print_usage(void)
{

	printf(_("Usage:"));
	printf(" %s -H <hostname> -L <log_directory> -C <counter> | list\n", progname);
	printf("   [ -w <threshold> ] [ -c <threshold> ]  [ -M <max_age> ]\n");
	printf("   [ -P <perfdata> ] [ -m ]\n");

}

